package cs214.webapp
package client

import java.net.{URLDecoder, URLEncoder}
import scala.concurrent.ExecutionContext.Implicits.global
import org.scalajs.dom
import org.scalajs.dom.{Element, KeyCode}
import org.scalajs.dom.html.{Input, Select, TextArea}
import scalatags.JsDom.all.*
import scalatags.JsDom.tags2.{article, section}

/** Parent class to web pages */
abstract class Page:
  val classList: String
  def renderInto(target: dom.Element): Unit

  def pageHeader(subtitle: String) =
    frag(h1(cls := "title", "ScalApp"), h2(subtitle))

  def replaceChildren(target: dom.Element)(frag: Frag): Unit =
    target.replaceChildren(frag.render)
    target.setAttribute("class", classList)

  private def path: Seq[String] =
    this match
      case HomePage =>
        Nil
      case InstanceCreationPage(appId) =>
        List("app", appId)
      case UIPage(appId, instanceId) =>
        List("app", appId, instanceId)
      case JoinPageLoader(appId, uiId, instanceId) =>
        List("app", appId, uiId, instanceId)
      case AppPage(appId, uiId, instanceId, userId) =>
        List("app", appId, uiId, instanceId, userId)
      case _ =>
        throw IllegalArgumentException(f"No path for $this!")

  def url =
    f"/${path.map(s => URLEncoder.encode(s, "UTF-8")).mkString("/")}"

object Page:
  /** Creates a component from a URL path
    * @param appInfos
    *   retrieves the app information given its id
    */
  def from(url: String): Page =
    val components = url.stripPrefix("/").stripSuffix("/").split("/")
    val decoded = components.map(c => URLDecoder.decode(c, "UTF-8")).toList
    decoded match
      case List("") =>
        HomePage
      case List("app", appId) =>
        InstanceCreationPage(appId)
      case List("app", appId, uiId) =>
        UIPage(appId, uiId)
      case List("app", appId, uiId, instanceId) =>
        JoinPageLoader(appId, uiId, instanceId)
      case List("app", appId, uiId, instanceId, userId) =>
        AppPage(appId, uiId, instanceId, userId)
      case _ =>
        throw IllegalArgumentException(f"Unknown url $url!")

/** The initial home page which fetches the list of apps */
object HomePage extends Page:
  val classList = "HomePage"

  def renderInto(target: Element) =
    Requests.listApps.map: response =>
      AppSelectionPage(response.apps).renderInto(target)

/** The app selection menu, where the user can create a new app. */
case class AppSelectionPage(apps: Seq[AppInfo]) extends Page:
  val classList = "AppSelectionPage"

  def selectApp(appId: AppId): Unit =
    WebClient.navigateTo(InstanceCreationPage(appId))

  def renderInto(target: Element) = replaceChildren(target):
    frag(
      pageHeader("Select an app"),
      apps.groupBy(_.year).toList.map: (year, appsOfYear) =>
        article(cls := "app-list",
          h3(year.toString),
          section(cls := "app-grid",
            appsOfYear.sortBy(_.name).map: appInfo =>
              figure(
                onclick := (() => selectApp(appInfo.id)),
                img(src := s"/static/${appInfo.id}.png"),
                figcaption(
                  h4(appInfo.name),
                  p(appInfo.description)
                )
              )
          )
      )
    )

/** The instance creation menu, where you enter the player names */
case class InstanceCreationPage(appId: AppId) extends Page:
  val classList = "InstanceCreationPage"

  def submit(e: dom.Event): Unit =
    e.preventDefault()
    val userIdStr = getElementById[Input]("user-ids").value
    val userIds = userIdStr.split("[;,]").map(_.strip).to(Seq).distinct
    Requests.createInstance(appId, userIds).map: resp =>
      WebClient.navigateTo(UIPage(appId, resp.instanceId))

  def renderInto(target: Element) = replaceChildren(target):
    frag(
      pageHeader(s"Create a new $appId instance"),
      form(
        onsubmit := submit,
        div(
          cls := "grid-form",
          label(`for` := "user-ids", "User IDs: "),
          input(
            `type` := "text",
            id := "user-ids",
            placeholder := "user1; user2; …",
            required := true,
            autofocus := true
          )
        ),
        input(`type` := "submit", value := "Start!")
      )
    )

/** The ui choice menu */
case class UIPage(appId: AppId, instanceId: InstanceId) extends Page:
  val classList = "UIPage"

  private val appUIs = WebClient.getApps(appId)

  private def cssId(idx: Int) = f"ui-$idx"

  private val handleKeyboardEvent: Function1[dom.KeyboardEvent, Unit] = (e: dom.KeyboardEvent) =>
    if e.keyCode == KeyCode.Enter then
      e.preventDefault()
      joinPageLoader()

  private def handleFormSubmission(e: dom.Event): Unit =
    e.preventDefault()
    joinPageLoader()

  private def getSelected =
    appUIs.zipWithIndex.find { (ui, i) =>
      getElementById[Input](cssId(i)).checked
    }.map(_._1)

  private def joinPageLoader(): Unit =
    getSelected.map: ui =>
      WebClient.navigateTo(JoinPageLoader(appId, ui.uiId, instanceId))

  def renderInto(target: Element) =
    require(appUIs.nonEmpty, f"No UI found for app with id $appId.")
    if appUIs.size == 1 then
      WebClient.navigateTo(JoinPageLoader(appId, appUIs(0).uiId, instanceId), overwriteHistory = true)
    else replaceChildren(target):
      dom.window.addEventListener("keydown", handleKeyboardEvent)
      frag(
        pageHeader("Choose your UI"),
        form(
          onsubmit := handleFormSubmission,
          fieldset(
            legend("Select your UI"),
            for (appUI, idx) <- appUIs.zipWithIndex
            yield div(
              input(
                `type` := "radio",
                id := cssId(idx),
                if idx == 0 then checked := "checked" else frag(),
                name := "ui",
                value := appUI.uiId,
                required := true
              ),
              label(`for` := cssId(idx), appUI.uiId)
            )
          ),
          input(`type` := "submit", value := "Play!")
        )
      )

/** The pre-connection menu, which fetches the user list. */
case class JoinPageLoader(appId: AppId, uiId: UIId, instanceId: InstanceId) extends Page:
  val classList = "JoinPageLoader"
  def renderInto(target: Element) =
    Requests.instanceInfo(instanceId).map: resp =>
      JoinPage(appId, uiId, instanceId, resp.userIds).renderInto(target)

/** The connection menu, where a user joins an existing app. */
case class JoinPage(appId: AppId, uiId: UIId, instanceId: InstanceId, userIds: Seq[UserId])
    extends Page:
  require(userIds.size > 0)
  val classList = "JoinPage"

  private def cssId(idx: Int) = f"user-$idx"

  private val handleKeyboardEvent: Function1[dom.KeyboardEvent, Unit] = (e: dom.KeyboardEvent) =>
    if e.keyCode == KeyCode.Enter then
      e.preventDefault()
      joinAppInstance()

  private def handleFormSubmission(e: dom.Event): Unit =
    e.preventDefault()
    joinAppInstance()

  private def getSelected =
    userIds.zipWithIndex.find { (u, i) =>
      getElementById[Input](cssId(i)).checked
    }.map(_._1)

  private def joinAppInstance(): Unit =
    getSelected.map: userId =>
      WebClient.navigateTo(AppPage(appId, uiId, instanceId, userId))

  def renderInto(target: Element): Unit =
    if userIds.size == 1 then
      WebClient.navigateTo(AppPage(appId, uiId, instanceId, userIds(0)), overwriteHistory = true)
    else replaceChildren(target):
      dom.window.addEventListener("keydown", handleKeyboardEvent)
      frag(
        pageHeader("Join app instance"),
        form(
          onsubmit := handleFormSubmission,
          fieldset(
            legend("Select your username"),
            for (userId, idx) <- userIds.zipWithIndex
            yield div(
              input(
                `type` := "radio",
                id := cssId(idx),
                if idx == 0 then checked := "checked" else frag(),
                name := "user",
                value := userId,
                required := true
              ),
              label(`for` := cssId(idx), userId)
            )
          ),
          input(`type` := "submit", value := "Join!")
        )
      )

/** The actual app content page */
case class AppPage(appId: AppId, uiId: UIId, instanceId: InstanceId, userId: UserId) extends Page:
  val classList = f"app $appId"

  private val app: ClientApp =
    val apps = WebClient.getApps(appId)
    require(
      !apps.isEmpty,
      s"Could not find an app UI corresponding to app id \"$appId\". There may be a " +
        s"mismatch between the client UI's appId and server logic's appId, or you may have " +
        s"forgotten to annotate your UI with @JSExportTopLevel(\"<yourAppId>\")"
    )
    val maybeApp = apps.find(_.uiId == uiId)
    require(
      maybeApp.isDefined,
      s"Could not find an app UI with id \"$uiId\"."
    )
    maybeApp.get

  def renderInto(target: Element) =
    replaceChildren(target):
      frag(header(id := "banner"), tag("section")(id := "app"))
    Requests.instanceInfo(instanceId).map: appInfo =>
      val port = dom.window.location.port
      val hostName = dom.window.location.hostname
      val protocol = if dom.window.location.protocol.startsWith("https") then "wss" else "ws"
      val endpoint = appInfo.wsEndpoint
        .replace("{{protocol}}", protocol)
        .replace("{{hostName}}", hostName + (if port.nonEmpty then f":$port" else ""))
        .replace("{{userId}}", URLEncoder.encode(userId, "UTF-8"))
      IpBanner(appInfo.shareUrl).renderInto(target.querySelector("#banner"))
      app.init(instanceId, userId, endpoint, target.querySelector("#app"))

/** The top-most share banner */
case class IpBanner(shareUrl: String) extends Page:
  val classList = "IpBanner"

  def renderInto(target: Element) =
    replaceChildren(target):
      val copyToClipboardButtonText = "📋 Copy to clipboard."
      frag(
        span("Share ", a(href := shareUrl, shareUrl), " with your friends to let them join!"),
        div(
          id := "banner-actions",
          button(
            onclick := (() => WebClient.navigateTo(HomePage)),
            "🏠 Back to home page"
          ),
          button(
            id := "copy-share-url",
            onclick := (() =>
              dom.window.navigator.clipboard.writeText(shareUrl)
              dom.document.getElementById("copy-share-url").innerText = "Copied!"

              dom.window.setTimeout(
                () => dom.document.getElementById("copy-share-url").innerText = copyToClipboardButtonText,
                1000
              )
            ),
            copyToClipboardButtonText
          )
        )
      )

def getElementById[T](id: String) =
  dom.document.getElementById(id).asInstanceOf[T]
