package cs214.webapp
package client

import scala.collection.mutable
import scala.util.{Failure, Success}
import scala.concurrent.{Future, Promise}

import org.scalajs.dom

type Target = dom.Element

abstract class WSClientApp extends ClientApp:
  WebClient.register(this)

  protected def init(userId: UserId, sendMessage: ujson.Value => Unit, target: Target): ClientAppInstance

  def init(instanceId: InstanceId, userId: UserId, endpoint: String, target: Target): ClientAppInstance =
    val socket = new dom.WebSocket(endpoint)
    socket.onopen = (event: dom.Event) => println("WebSocket connection opened")
    socket.onclose = (event: dom.CloseEvent) => println(s"WebSocket connection closed: ${event.reason}")
    socket.onerror = (event: dom.Event) => println(s"WebSocket error: ${event.`type`}")

    val sendMessage = (js: ujson.Value) => socket.send(js.toString)
    val client = init(userId, sendMessage, target)
    socket.onmessage = msg =>
      val js = ujson.read(msg.data.toString)
      client.onMessage(SocketResponseWire.decode(js).flatten)

    client

/** Instance of a client-side state machine application.
  *
  * @param userId
  *   the user for which we're creating a client application
  * @param sendMessage
  *   how to send a json message
  */
abstract class StateMachineClientAppInstance[Event, View](
    userId: UserId,
    sendMessage: ujson.Value => Unit,
)(
    using scala.concurrent.ExecutionContext
) extends ClientAppInstance, ClientUI[View]:
  /** Provides serialization methods for events and views */
  def wire: AppWire[Event, View]

  protected def sendEvent(event: Event) =
    sendMessage(wire.eventFormat.encode(event))

  /** @inheritdoc */
  override def onMessage(msg: util.Try[ujson.Value]): Unit =
    msg.flatMap(EventResponse.Wire.decode) match
      case Failure(msg)              => WebClient.crash(Exception(msg))
      case Success(Failure(msg))     => displayError(msg.getMessage)
      case Success(Success(actions)) => actionsQueue.enqueueAll(actions)
    tryProcessActionsQueue()

  /** Actions waiting to be processed. */
  private val actionsQueue: mutable.Queue[Action[ujson.Value]] = mutable.Queue()
  private var processing = false

  private def tryProcessActionsQueue(): Unit =
    if !processing then processActions()

  private def processActions(): Unit =
    require(!processing)
    def loop: Unit =
      processing = actionsQueue.nonEmpty
      if processing then
        processAction(actionsQueue.dequeue()).andThen(_ => loop)
    loop

  /** Execute a single action sent by the server. */
  private def processAction(jsonAction: Action[ujson.Value]): Future[Unit] =
    jsonAction match
      case Action.Alert(msg) =>
        displayAlert(msg)
        after(0)
      case Action.Pause(durationMs) =>
        after(durationMs)
      case Action.Render(js) =>
        wire.viewFormat.decode(js) match
          case Failure(exception) => displayError(exception.getMessage)
          case Success(view) => displayView(view)
        after(0)
