package cs214.webapp
package client
package graphics

import scala.util.{Try, Success, Failure}
import scala.concurrent.Future

import scala.scalajs.js
import scala.scalajs.concurrent.JSExecutionContext.Implicits.queue
import org.scalajs.dom

import scalatags.JsDom.Frag

trait PatchableUI:
  /** Updates the DOM Element with a new fragment. */
  def patchUI(target: dom.Element, fr: Frag): Unit

/** Updates the DOM using a custom library instead of `replaceChildren`.
  *
  * This avoid flickering and losing UI state such as the current text
  * selection, form inputs, CSS transitions, etc.
  *
  **/
trait DomPatchingUI extends PatchableUI:
  protected type PatchFn = (dom.Element, dom.Node) => Unit

  protected def importLib(name: String, url: String): Future[js.Dynamic] =
    js.`import`[js.Dynamic](url).toFuture.recover: exn =>
      println(f"Exception while loading `$name`: ${exn.getMessage()}")
      println(f"`$name` library loading failed: falling back to `replaceChildren` permanently")
      throw exn

  val patchLib: Future[PatchFn] // Fetch the dom-patching library

  override def patchUI(target: dom.Element, fr: Frag): Unit =
    patchLib.value match
      case Some(Success(patchFn)) =>
        val wrapper = target.cloneNode(deep = false)
        wrapper.appendChild(fr.render)
        patchFn(target, wrapper)
      case Some(Failure(_)) =>
        target.replaceChildren(fr.render)
      case None =>
        println("UI library not loaded yet: falling back to `replaceChildren` temporarily.")
        target.replaceChildren(fr.render)

/** Updates the DOM with morphdom (does not support events).
  *
  * See https://github.com/patrick-steele-idem/morphdom for details. */
trait MorphdomPatchingUI extends DomPatchingUI:
  override val patchLib =
    importLib("morphdom", "https://cdn.jsdelivr.net/npm/morphdom@2.7.4/+esm").map: m =>
      (el, n) => m.default(el, n)

/** Updates the DOM with nanomorph.
  *
  * See https://github.com/choojs/nanomorph for details. */
trait NanomorphPatchingUI extends DomPatchingUI:
  override val patchLib =
    importLib("nanomorph", "https://cdn.jsdelivr.net/npm/nanomorph@5.4.3/+esm").map: m =>
      (el, n) => m.default(el, n)

// Default to Nanomorph
trait IncrementalUI extends NanomorphPatchingUI
