package cs214.webapp.client
package graphics

import cs214.webapp.UserId
import cs214.webapp.client.StateMachineClientAppInstance
import org.scalajs.dom
import scalatags.JsDom.all.*

enum HTMLAttribute:
  case cls
  case style

type CSSProperties = Map[String, String]
extension (cssProps: CSSProperties)
  def toCSS(): String = cssProps
    .map((k, v) => f"${k}: ${v}")
    .mkString("; ")

enum MouseEvent:
  case Click(button: MouseButton)
  case MouseDown(button: MouseButton)
  case MouseUp(button: MouseButton)
  case HoverEnter
  case HoverLeave

/** The console is segmented with [[TextSegment]]s. Each segment represents a
  * piece of text with its own modifiers and event listeners.
  *
  * If multiple parts of your text need to respond independently to mouse
  * events or need a different style, you should consider separating them in
  * different [[TextSegment]]s
  *
  * @param text
  *   The text to be displayed
  * @param onMouseEvent
  *   How to handle mouse events happening on this text segment
  * @param cssProperties
  *   The css properties to inline with this text segment (html style
  *   attribute)
  * @param htmlAttributes
  *   The html attributes to add to this text segment
  * @param modifiers
  *   Any additional `JsDom.Modifier`s you may want to add to the element
  */
case class TextSegment(
    text: String,
    onMouseEvent: Option[MouseEvent => Unit] = None,
    cssProperties: CSSProperties = Map(),
    htmlAttributes: Map[HTMLAttribute, String] = Map(),
    modifiers: Modifier*
):
  def toFrag: Frag =
    val css = cssProperties.toCSS()
    span(
      text,
      conditionalModifier(HTMLAttribute.style, style, if css.isEmpty then None else Some(css)),
      conditionalModifier(HTMLAttribute.cls, cls, None),
      modifiers,
      bindMouseEvents(onMouseEvent)
    )

  private def bindMouseEvents[T](handler: Option[MouseEvent => Unit]): Modifier =
    handler match
      case None => frag()
      case Some(fn) => Seq(
        onclick := { (e: dom.MouseEvent) => fn(MouseEvent.Click(MouseButton.from(e.button))) },
        onmousedown := { (e: dom.MouseEvent) => fn(MouseEvent.MouseDown(MouseButton.from(e.button))) },
        onmouseup := { (e: dom.MouseEvent) => fn(MouseEvent.MouseUp(MouseButton.from(e.button))) },
        onmouseover := { (e: dom.MouseEvent) => fn(MouseEvent.HoverEnter) },
        onmouseout := { (e: dom.MouseEvent) => fn(MouseEvent.HoverLeave) }
      )

  private def conditionalModifier(attr: HTMLAttribute, target: Attr, default: => Option[String] = None): Modifier =
    htmlAttributes.get(attr).orElse(default) match
      case None => frag()
      case Some(v) => target := v


/** Text-based UI. Detects mouse events and has an integrated text field. */
abstract class TextClientAppInstance[Event, View](
    userId: UserId,
    sendMessage: ujson.Value => Unit,
    target: Target
) extends WebClientAppInstance[Event, View](userId, sendMessage, target):
  /** Transforms a text input from the user into an optional app [[Event]]
    * @param view
    *   The current view
    * @param text
    *   The input text typed by the user
    * @return
    *   An optional application event
    */
  def handleTextInput(view: View, text: String): Option[Event]

  /** Transforms a view received by the server into a segmented console display.
    * Each [[TextSegment]] responds independently to mouse events and defines
    * its own style.
    *
    * @param userId
    *   The id of the user using this client
    * @param view
    *   The view of the application received by the server
    * @return
    *   The different text segments of the console
    */
  def renderView(userId: UserId, view: View): Vector[TextSegment]

  override protected def render(userId: UserId, view: View): Frag =
    currView = Some(view)

    frag(
      div(
        cls := "textapp",
        div(
          cls := "console",
          pre(renderView(userId, view).map(_.toFrag))
        ),
        input(
          `type` := "text",
          autofocus := true,
          id := "console-input",
          tabindex := 0,
          onkeydown := handleKeyDown,
        )
      )
    )

  private var currView: Option[View] = None

  protected def focusConsoleInput(): Unit =
    dom.document.getElementById("console-input").asInstanceOf[dom.HTMLInputElement].focus()

  private def handleKeyDown(e: dom.KeyboardEvent): Unit = currView
    .flatMap(view =>
      if e.keyCode == dom.KeyCode.Enter then
        e.target match
          case inputElement: dom.HTMLInputElement =>
            e.preventDefault()
            val v = inputElement.value
            inputElement.value = ""
            handleTextInput(view, v)
      else
        None
    ).foreach(sendEvent)

  override def css: String = super.css +
    """
    | .textapp {
    |   font-size: 1rem;
    | }
    |
    | .textapp input {
    |   height: 1em;
    |   width: 100%;
    |   border: solid 1px black;
    |   box-sizing: border-box;
    | }
    |
    | .textapp .console {
    |     overflow: hidden;
    |     padding: 0.5em;
    |     font-size: 1em;
    |     display: flex;
    |     flex-direction: column;
    |     justify-content: center;
    |     align-items: start;
    |     border: solid 1px black;
    | }
    |
    | .textapp .console pre {
    |   white-space: pre-wrap;
    |   margin: 0;
    | }
    |
    | .textapp .console span {
    |     box-sizing: border-box;
    | }
    """.stripMargin
