package cs214.webapp
package server

import scala.util.Try

/** A state machine describes the core logic of an application.
  *
  * @tparam Event
  *   The type of the events that this application expects.
  * @tparam State
  *   The type of internal states of this application.
  * @tparam View
  *   The type of the views sent to the clients.
  */
abstract class StateMachine[Event, State, View]:
  /** An identifier for this app type. */
  def appInfo: cs214.webapp.AppInfo

  /** Provides serialization methods for events and views. */
  def wire: AppWire[Event, View]

  /** Initializes a new application. */
  def init(clients: Seq[UserId]): State

  /** Simulates a transition of the state machine.
    *
    * @param state
    *   The current [[State]] of the application.
    * @param userId
    *   The [[UserId]] of the user triggering the event
    * @param event
    *   The [[Event]] received
    * @return
    *   A sequence of commands to be sent to clients, or [[scala.util.Failure]]
    *   if the event is illegal given the current state.
    */
  def transition(state: State)(userId: UserId, event: Event): Try[Seq[Action[State]]]

  /** Projects an application state to produce a user-specific view.
    *
    * This function hides any details of the application's state that the user
    * should not have access to.
    *
    * @param state
    *   A [[State]] of the application.
    * @param userId
    *   The [[UserId]] of a user.
    * @return
    *   A view for user [[uid]].
    */
  def project(state: State)(userId: UserId): View

/** Extend this class if you want your state machine to receive a [[Tick]] event
  * in addition to your own custom events. This event will be received every
  * [[clockPeriodMs]] milliseconds
  *
  * @tparam Event
  *   The type of the events that this application expects.
  * @tparam State
  *   The type of internal states of this application.
  * @tparam View
  *   The type of the views sent to the clients.
  */
abstract class ClockDrivenStateMachine[Event, State, View]
    extends StateMachine[Either[Tick, Event], State, View]:

  /** The [[Tick]] event will be received by the state machine every
    * [[clockPeriodMs]] milliseconds.
    */
  val clockPeriodMs: Int

  /** Provides serialization methods for events and views of a
    * [[ClockDrivenStateMachine]].
    */
  val clockDrivenWire: AppWire[Event, View]

  override object wire extends ClockDrivenWire(clockDrivenWire)

private class ClockDrivenWire[Event, View](wire: AppWire[Event, View]) extends AppWire[Either[Tick, Event], View]:
  override object eventFormat extends WireFormat[Either[Tick, Event]]:
    override def decode(json: ujson.Value): Try[Either[Tick, Event]] = Try:
      Right(wire.eventFormat.decode(json).get)

    override def encode(t: Either[Tick, Event]): ujson.Value =
      require(t.isRight, "Tick events should not be sent over the network.")
      wire.eventFormat.encode(t.toOption.get)

  override val viewFormat: WireFormat[View] = wire.viewFormat
