package cs214.webapp
package server
package web

import scala.collection.mutable
import scala.util.{Success, Failure, Try}
import scala.concurrent.duration.Duration

import cask.endpoints.WsChannelActor
import java.util.concurrent.atomic.AtomicBoolean

case class BroadcastException(ex: Throwable) extends Throwable

/** Server-side apps definition and abstractions */

/** A server-side app factory, ready to create new `ServerApp` instances. */
private[web] sealed trait ServerAppFactory:
  def appInfo: AppInfo
  def init(instanceId: InstanceId, registeredUserIds: Seq[UserId]): ServerApp

private[web] type PendingActions = Map[UserId, Seq[Action[ujson.Value]]]

private[web] case class StateMachineServerAppFactory[E, S, V](sm: StateMachine[E, S, V]) extends ServerAppFactory:
  override def appInfo: AppInfo = sm.appInfo
  override def init(instanceId: InstanceId, registeredUserIds: Seq[UserId]): ServerApp =
    StateMachineServerApp(sm, instanceId, registeredUserIds)

private[web] case class ClockDrivenStateMachineServerAppFactory[E, S, V](sm: ClockDrivenStateMachine[E, S, V]) extends ServerAppFactory:
  override def appInfo: AppInfo = sm.appInfo
  override def init(instanceId: InstanceId, registeredUserIds: Seq[UserId]): ServerApp =
    ClockDrivenStateMachineServerApp(sm, instanceId, registeredUserIds)

private[web] abstract class ServerApp:
  val instanceId: InstanceId
  val registeredUserIds: Seq[UserId]

  protected object instanceLock

  def appInfo: AppInfo

  /** Compute setup actions to send to client. */
  def respondToNewClient(userId: UserId): ujson.Value

  /** Transition from the current state of the application to the next.
    *
    * @param clients
    *   The set of clients currently connected in the application.
    * @param userId
    *   The id of the user who triggered the event.
    * @param eventJs
    *   The serialised event triggered by the user.
    * @return
    *   A map which links each user to a queue of views they should receive.
    */
  def transition(clients: Seq[UserId])(userId: UserId, eventJs: ujson.Value): Try[PendingActions]

  private val channels =
    mutable.Map[UserId, mutable.Set[cask.WsChannelActor]]().withDefault(_ => mutable.Set())

  def connect(userId: UserId)
             (implicit cc: castor.Context, log: cask.Logger): cask.WebsocketResult = instanceLock.synchronized:
    cask.WsHandler: channel =>
      channels.getOrElseUpdate(userId, mutable.Set()).add(channel)
      println(f"[${appInfo.id}/$instanceId] client \"$userId\" connected")
      send(userId):
        val view = respondToNewClient(userId)
        EventResponse.Wire.encode(Success(List(Action.Render(view))))
      cask.WsActor {
        case cask.Ws.Error(e) =>
          throw e
        case cask.Ws.Close(code, reason) =>
          disconnect(userId, channel)
        case cask.Ws.Text(data) =>
          handleMessage(userId, ujson.read(data))
      }

  def disconnect(userId: UserId, channel: WsChannelActor): Unit = instanceLock.synchronized:
    channels(userId).remove(channel)
    if channels(userId).isEmpty then channels.remove(userId)
    println(f"[${appInfo.id}/$instanceId] client \"$userId\" disconnected")

  /** Enumerates clients connected to this instance. */
  def connectedClients: Seq[UserId] = instanceLock.synchronized:
    channels.keysIterator.toSeq

  def shutdown() = instanceLock.synchronized:
     for channels <- channels.values
         channel <- channels
     do
       channel.send(cask.Ws.Close())

  /** Sends a message to a specific client. */
  private def send(userId: UserId)(message: ujson.Value): Unit = instanceLock.synchronized:
    val wrapped = SocketResponseWire.encode(util.Success(message)).toString
    for channel <- channels(userId) do
      channel.send(cask.Ws.Text(wrapped))

  def handleMessage(userId: UserId, msg: ujson.Value): Unit = instanceLock.synchronized:
    announceTransition(userId):
      transition(connectedClients)(userId, msg)

  protected def announceTransition(userId: UserId)(result: Try[PendingActions]): Unit = instanceLock.synchronized:
    val pending: Map[String, Try[Seq[Action[ujson.Value]]]] = result match
      case Failure(BroadcastException(ex)) =>
        connectedClients.iterator.map(_ -> Failure(ex)).toMap
      case Failure(ex) =>
        Map(userId -> Failure(ex))
      case Success(pending) =>
        pending.view.mapValues(Success.apply).toMap
    for (userId, actions) <- pending do
      send(userId):
        EventResponse.Wire.encode(actions)

private[web] class StateMachineServerApp[E, S, V](
    sm: StateMachine[E, S, V],
    val instanceId: InstanceId,
    val registeredUserIds: Seq[UserId],
) extends ServerApp:
  private var state = sm.init(registeredUserIds)

  private def project(st: S)(userId: UserId): ujson.Value =
    sm.wire.viewFormat.encode(sm.project(st)(userId))

  private def project(actions: Seq[Action[S]])(userId: UserId): Seq[Action[ujson.Value]] =
    actions.map(_.map(st => project(st)(userId)))

  override def appInfo = sm.appInfo

  override def respondToNewClient(userId: UserId): ujson.Value =
    project(state)(userId)

  override def transition(clients: Seq[UserId])(userId: UserId, eventJs: ujson.Value): Try[PendingActions] = Try:
    transition(clients)(userId, sm.wire.eventFormat.decode(eventJs).get).get

  def transition(clients: Seq[UserId])(userId: UserId, event: E): Try[PendingActions] = Try:
    if WebServer.debug then
      println(f"[debug] got event: $event")

    val actions = sm.transition(state)(userId, event).get

    if WebServer.debug then
      println(s"[debug] transition produced actions ${actions.mkString(", ")}")

    val clientActions = clients.map: userId =>
      userId -> actions.map(_.map(st => sm.project(st)(userId)))

    val pending = Map.from:
      clients.iterator.map: userId =>
        userId -> project(actions)(userId)

    // Update state last, so that any error cancels the transition
    state = actions.collect { case Action.Render(t) => t }.lastOption.getOrElse(state)

    pending

private final class Clock(val name: String, periodMs: Long)(action: => Unit) extends Thread(name):
  private val running = AtomicBoolean(true)

  override def run(): Unit =
    require(running.compareAndExchange(false, true))
    println(f"[$name] started")
    while running.get() do
      action
      Thread.sleep(periodMs)
    println(f"[$name] stopped")

  def shutdown() =
    running.set(false)

private[web] final class ClockDrivenStateMachineServerApp[E, S, V](
    sm: ClockDrivenStateMachine[E, S, V],
    instanceId: InstanceId,
    registeredUserIds: Seq[UserId],
) extends StateMachineServerApp(sm, instanceId, registeredUserIds):
  private val clock =
    val name = f"${appInfo.id}/$instanceId/clock"
    Clock(name, sm.clockPeriodMs)(handleClockTick)

  clock.start()

  def handleClockTick: Unit = instanceLock.synchronized:
    try
      val event = Left(Tick(System.currentTimeMillis))
      announceTransition(clock.name):
        transition(connectedClients)(clock.name, event).recover: ex =>
          throw(BroadcastException(ex))
    catch ex =>
      println(f"[${appInfo.id}/$instanceId] Uncaught exception in clock action: $ex")

  override def shutdown(): Unit =
    clock.shutdown()
    super.shutdown()
