package cs214.webapp
package server
package web

import scala.collection.{concurrent, mutable, immutable}
import scala.util.{Failure, Success, Try}
import cask.endpoints.WsChannelActor

private case class Wordlist(fname: String):
  val words: Seq[String] =
    import scala.io.{Source, Codec}
    val wordstream =
      Option(this.getClass.getResourceAsStream(f"/$fname"))
        .getOrElse(sys.error(f"File not found: $fname"))
    try
      Source.fromInputStream(wordstream)(Codec.UTF8)
        .getLines().map(_.split("\t").last).to(Vector)
    finally wordstream.close()

  def randomWord: String =
    import scala.util.Random
    words(Random.nextInt(words.size))

/** Contains the web server state and functionalities */
object WebServer:
  private val instanceIdsFileName = "eff_short_wordlist_1.txt" // Where to find the list of possible instance ids
  private val instanceIdLength = 4 // How many words do we concatenate for the instance id
  private val instanceIdWords = Wordlist(instanceIdsFileName)

  private[web] val debug = false

  /** Mapping from app ids to their app class */
  private[web] val appDirectory: mutable.Map[AppId, ServerAppFactory] = mutable.Map()

  /** Mapping from app instance ids to their running instances */
  private[web] val instances: concurrent.Map[InstanceId, ServerApp] =
    concurrent.TrieMap()

  /** Registers the given app */
  private def register(app: ServerAppFactory): Unit =
    require(!appDirectory.contains(app.appInfo.id))
    appDirectory.put(app.appInfo.id, app)
    println(f"[${app.appInfo.id}] registered")

  /** Registers the given state-machine based app. */
  def register[E, V, S](sm: StateMachine[E, V, S]): Unit = sm match
    case csm: ClockDrivenStateMachine[_, _, _] =>
      register(ClockDrivenStateMachineServerAppFactory(csm))
    case _ =>
      register(StateMachineServerAppFactory(sm))

  /** Computes an unused instance ID and calls `body` with it. */
  private def withFreshInstanceId[T](body: InstanceId => T): T =
    instances.synchronized:
      def freshId = Iterator.fill(instanceIdLength)(instanceIdWords.randomWord).mkString("-")
      val instanceId = Iterator.continually(freshId).filterNot(instances.contains).next()
      body(instanceId)

  /** Creates a new instance of `appId`. */
  private[web] def createInstance(appId: AppId, registeredUserIds: Seq[UserId]): InstanceId =
    withFreshInstanceId: instanceId =>
      instances(instanceId) = appDirectory(appId).init(instanceId, registeredUserIds)
      println(f"[$appId] instance created $instanceId")
      instanceId

  private[web] def shutdownApp(instanceId: InstanceId): Unit =
    instances.remove(instanceId).map: app =>
      app.shutdown()
      println(s"[${app.appInfo.id}][$instanceId] shut down")
