package cs214.webapp

import ujson.*

import java.util.UUID
import scala.util.{Failure, Success, Try}

/** An action that client should perform */
enum Action[+StateOrView]:
  /** Show a message box. */
  case Alert(msg: String)

  /** Wait before processing the next action. */
  case Pause(durationMs: Int)

  /** Redraw the UI with new data. */
  case Render(st: StateOrView)

  def map[T](f: StateOrView => T): Action[T] =
    this match
      case Alert(msg)        => Alert(msg)
      case Pause(durationMs) => Pause(durationMs)
      case Render(st)        => Render(f(st))

object Action:
  object Wire extends WireFormat[Action[Value]]:
    def encode(t: Action[Value]): Value =
      t match
        case Alert(msg)        => Obj("type" -> "Alert", "msg" -> msg)
        case Pause(durationMs) => Obj("type" -> "Pause", "durationMs" -> durationMs)
        case Render(js)        => Obj("type" -> "Render", "js" -> js)

    def decode(json: Value): Try[Action[Value]] = Try:
      json("type").str match
        case "Alert"  => Alert(json("msg").str)
        case "Pause"  => Pause(json("durationMs").num.toInt)
        case "Render" => Render(json("js"))
        case _        => throw DecodingException(f"Unexpected action: $json")

/** A response to an event */
type EventResponse = Try[Seq[Action[ujson.Value]]]
object EventResponse:
  val Wire: WireFormat[EventResponse] = TryWire(SeqWire(Action.Wire))

/** A response to a socket message */
type SocketResponse = Try[ujson.Value]
val SocketResponseWire = TryWire(IdentityWire)

/** A response to the list-apps query */
case class ListAppsResponse(apps: Seq[AppInfo])

object ListAppsResponse:
  object Wire extends WireFormat[ListAppsResponse]:
    def encode(t: ListAppsResponse): ujson.Value =
      ujson.Obj("apps" -> Arr(t.apps.map {
        case AppInfo(id, name, description, year) =>
          Obj("id" -> id, "name" -> name, "desc" -> description, "year" -> year)
      }*))
    def decode(js: ujson.Value): Try[ListAppsResponse] = Try:
      ListAppsResponse(
        js("apps").arr
          .map(_.obj)
          .map(js =>
            AppInfo(
              id = js("id").str,
              name = js("name").str,
              description = js("desc").str,
              year = js("year").num.toInt
            )
          )
          .to(Seq)
      )

/** An HTTP request sent to create a new application instance for the server */
case class CreateInstanceRequest(appName: String, userIds: Seq[UserId])

object CreateInstanceRequest:
  object Wire extends WireFormat[CreateInstanceRequest]:
    val strsWire = SeqWire(StringWire)
    def encode(t: CreateInstanceRequest): ujson.Value =
      ujson.Obj("appName" -> Str(t.appName), "userIds" -> strsWire.encode(t.userIds))
    def decode(js: ujson.Value): Try[CreateInstanceRequest] = Try:
      CreateInstanceRequest(js("appName").str, strsWire.decode(js("userIds")).get)

case class CreateInstanceResponse(instanceId: InstanceId)

object CreateInstanceResponse:
  object Wire extends WireFormat[CreateInstanceResponse]:
    def encode(t: CreateInstanceResponse): ujson.Value =
      ujson.Obj("instanceId" -> Str(t.instanceId))
    def decode(js: ujson.Value): Try[CreateInstanceResponse] = Try:
      CreateInstanceResponse(js("instanceId").str)

/** The response to an AppInfo query */
case class InstanceInfoResponse(instanceId: InstanceId, userIds: Seq[UserId], wsEndpoint: String, shareUrl: String)

object InstanceInfoResponse:
  val strsWire = SeqWire(StringWire)
  object Wire extends WireFormat[InstanceInfoResponse]:
    def encode(t: InstanceInfoResponse): ujson.Value =
      ujson.Obj(
        "instanceId" -> t.instanceId,
        "userIds" -> strsWire.encode(t.userIds),
        "wsEndpoint" -> t.wsEndpoint,
        "shareUrl" -> t.shareUrl
      )
    def decode(js: ujson.Value): Try[InstanceInfoResponse] = Try:
      InstanceInfoResponse(
        js("instanceId").str,
        strsWire.decode(js("userIds")).get,
        js("wsEndpoint").str,
        js("shareUrl").str
      )
